{
                               xineEngine.pas
                             -------------------
    copyright            : (C) 2002 by Ryan Rusaw
    email                : rrusaw@sasktel.net

 ***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
}

unit xineengine;

{$mode objfpc}{$H+}
interface

uses
  SysUtils, Classes, libc, xine, xineaudioplugin, xinevideoplugin, xineStream,Controls,
  xlib,gdk,pthreads,Forms,x;

const
  movie_brightness = 32767;
  movie_contrast = 32767;
  movie_hue = 32767;
  movie_saturation = 32767;
  movie_volume = 32767;

type


  PCharArray = array of PChar;

  TXineEvent = procedure(data:Pointer) of object;

  TxineUserData = record
     rect: x11_rectangle_t;
     old_width, old_height : Integer;
     resizeFactor : Double;
     displayRatio : Double;
     Display: PDisplay;
     shmEventId: Integer;
     { event callbacks }
     OnPlaybackFinished: TXineEvent;
     OnChannelChanged: TXineEvent;
     OnSetTitle: TXineEvent;
     OnProgress: TXineEvent;
     OnNumButtons: TXineEvent;
     OnMRLReference: TXineEvent;
     W:TWinControl;
  end;

  TMyThread = class;
  { TxineEngine }

  TxineEngine = class(TComponent)
  private
    { Private declarations }
    FOnPlaybackFinished: TXineEvent;
    FOnChannelChanged: TXineEvent;
    FOnSetTitle: TXineEvent;
    FOnProgress: TXineEvent;
    FOnNumButtons: TXineEvent;
    FOnMRLReference: TXineEvent;
    procedure setVideoPlugin(plugin: TxineVideoPlugin);
    procedure setAudioPlugin(plugin: TxineAudioPlugin);
    procedure setXStream(stream: TxineStream);
  protected
    { Protected declarations }
  public
    myVidPlug : TxineVideoPlugin;
    myAudPlug : TxineAudioPlugin;
    myXine : Pxine_t;
    myXStream : TxineStream;
    UserData : TxineUserData;
    queue: Pxine_event_queue_t;
    init:Boolean;
    Thread:TMyThread;

    { Public declarations }
    procedure Re;
    procedure Initialize;
    function Open(const filename: string):Integer;
    procedure Close;
    function Play:Integer;
    procedure Pause;
    procedure Stop;
    procedure Rewind;
    procedure FastForward;
    procedure Seek (pos: Integer);
    procedure Menu (id : Integer);
    procedure SetVolume(vol:Integer);
    procedure Exit;
    procedure SetProperty (prop, value:Integer);
    procedure SetAudioChannel (chan: Integer);
    procedure SetSubtitleChannel (chan: Integer);
    procedure SetZoom (zoom: Integer);
    function GetAutoplayPluginList:TStringList;
    function OpenAutoplayInputPluginList(str: String):TStringList;
    procedure SendInputEvent(button,x,y,typ: Integer);
    function PPCharToStringList(items: PPChar):TStringList;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; Override;
  published

    { Published declarations }
    property videoPlugin: TxineVideoPlugin read myVidPlug write setVideoPlugin;
    property audioPlugin: TxineAudioPlugin read myAudPlug write setAudioPlugin;
    property xStream: TxineStream read myXStream write setXStream;
    property OnPlaybackFinished: TXineEvent read FOnPlaybackFinished
          write FOnPlaybackFinished;
    property OnChannelChanged: TXineEvent read FOnChannelChanged
          write FOnChannelChanged;
    property OnSetTitle: TXineEvent read FOnSetTitle
          write FOnSetTitle;
    property OnProgress: TXineEvent read FOnProgress
          write FOnProgress;
    property OnNumButtons: TXineEvent read FOnNumButtons
          write FOnNumButtons;
    property OnMRLReference: TXineEvent read FOnMRLReference
          write FOnMRLReference;
  end;

  { TMyThread }
  TMyThread = class(TThread)
  private

  protected
    v:boolean;
  public
    xine:TxineEngine;
    procedure test;
    constructor Create(CreateSuspended: Boolean; const StackSize: SizeUInt=DefaultStackSize);
    destructor Destroy; override;
    procedure Execute; override;
  published

  end; // TMyThread
procedure Register;

procedure XineEventCB(user_data:Pointer; event: Pxine_event_t); cdecl;

implementation

procedure Register;
begin
  RegisterComponents('XINE', [TxineEngine]);
end;

procedure TxineEngine.setVideoPlugin(plugin: TxineVideoPlugin);
begin
     myVidPlug := plugin;
end;

procedure TxineEngine.setAudioPlugin(plugin: TxineAudioPlugin);
begin
     myAudPlug := plugin;
end;

procedure TxineEngine.setXStream(stream: TxineStream);
begin
     myXStream := stream;
end;

procedure TxineEngine.Re;
begin
  xine_event_dispose_queue(queue);
  xine_dispose(myXStream.GetStream);

  myXStream.SetStream(myXine, myAudPlug.AoDriver, myVidPlug.VoDriver);
  queue := xine_event_new_queue (myXStream.GetStream);
  xine_event_create_listener_thread (queue, @XineEventCB, @UserData);
end;

procedure TxineEngine.Initialize;
var p:pDisplay;
begin
  p:= XOpenDisplay(nil);
  myXine := xine_new();
  xine_config_load(myXine,strcat(getenv('HOME'),PChar('.xine/config')));
  xine_init(myXine);
  if myAudPlug <> NIL then
    myAudPlug.Initialize(myXine);

  if myVidPlug <> NIL then begin
    myVidPlug.Display := @p^;
    XLockDisplay(myVidPlug.Display);
    UserData.rect.x := 0;
    UserData.rect.y := 0;
    UserData.OnChannelChanged := OnChannelChanged;
    UserData.OnPlaybackFinished := OnPlaybackFinished;
    UserData.OnSetTitle := OnSetTitle;
    UserData.OnProgress := OnProgress;
          UserData.OnNumButtons := OnNumButtons;
          UserData.OnMRLReference := OnMRLReference;
          if ((myVidPlug.vWidget <> NIL) and (myVidPlug.IsEmbeddable)) then
          begin
               UserData.rect.w := myVidPlug.vWidget.Width;
               UserData.rect.h := myVidPlug.vWidget.Height;
               // set old values
               UserData.old_width := myVidPlug.vWidget.Width;
               UserData.old_height := myVidPlug.vWidget.Width;
               myVidPlug.Initialize(myXine, @UserData);
          end
          else
          begin
               UserData.rect.w := defaultWidth;
               UserData.rect.h := defaultHeight;
               // set old values
               UserData.old_width :=defaultWidth;
               UserData.old_height :=defaultHeight;
               myVidPlug.Initialize(myXine, @UserData);
          end;
     end;

     myXStream.SetStream(myXine, myAudPlug.AoDriver, myVidPlug.VoDriver);
     XUnlockDisplay (myVidPlug.Display);
     queue := xine_event_new_queue (myXStream.GetStream);
     xine_event_create_listener_thread (queue, @XineEventCB, @UserData);
   init:=True;
end;

function TxineEngine.Open(const filename: string):Integer;
var i:integer;
begin
  //  XLockDisplay (myVidPlug.Display);
   xine_set_param( myXStream.GetStream, XINE_PARAM_METRONOM_PREBUFFER, 0);
//   xine_usec_sleep(500);
   result:=xine_open(myXStream.GetStream,PChar(filename));
{  xine_set_param(myXStream.GetStream, XINE_PARAM_AV_OFFSET, 0);
  xine_set_param(myXStream.GetStream, XINE_PARAM_SPU_OFFSET, 0);
  xine_set_param (myXStream.GetStream, XINE_PARAM_VO_SATURATION, movie_saturation);
  xine_set_param (myXStream.GetStream, XINE_PARAM_VO_BRIGHTNESS, movie_brightness);
  xine_set_param (myXStream.GetStream, XINE_PARAM_VO_CONTRAST, movie_contrast);
  xine_set_param (myXStream.GetStream, XINE_PARAM_VO_HUE, movie_hue);}
//  xine_set_param( myXStream.GetStream, XINE_PARAM_VO_CROP_TOP, 5);
//xine_set_param( myXStream.GetStream, XINE_PARAM_VO_CROP_BOTTOM, 8);
//  xine_set_param( myXStream.GetStream, XINE_PARAM_VO_CROP_LEFT, 5);
//xine_set_param( myXStream.GetStream, XINE_PARAM_VO_CROP_RIGHT, 5);
//  xine_play(myXStream.GetStream,0,);

//  Sleep(400);
//xine_m
//  xine_usec_sleep(400);
  i:=xine_play(myXStream.GetStream,0,0);
  xine_set_param(myXStream.GetStream,XINE_PARAM_SPEED,XINE_SPEED_NORMAL);
  xine_set_param( myXStream.GetStream, XINE_PARAM_METRONOM_PREBUFFER, 9000);
  //  XUnlockDisplay (myVidPlug.Display);
  result:=0;
end;

procedure TxineEngine.Close;
begin
  xine_close(myXStream.GetStream);
end;

function TxineEngine.Play:Integer;
begin

//  pthread_detach(pthread_self);
  result:=xine_play(myXStream.GetStream,0,0);
end;

procedure TxineEngine.Pause;
begin
  xine_set_param(myXStream.GetStream,XINE_PARAM_SPEED,XINE_SPEED_PAUSE);
end;

procedure TxineEngine.Stop;
begin
  xine_stop(myXStream.GetStream);
end;

procedure TxineEngine.Rewind;
begin
     // not implemented;
end;

procedure TxineEngine.FastForward;
begin
  if (xine_get_param(myXStream.GetStream,XINE_PARAM_SPEED) = XINE_SPEED_FAST_2) then
  begin
    xine_set_param(myXStream.GetStream,XINE_PARAM_SPEED, XINE_SPEED_FAST_4);
  end
  else
  begin
     if (xine_get_param(myXStream.GetStream,XINE_PARAM_SPEED) = XINE_SPEED_FAST_4) then
     begin
        xine_set_param(myXStream.GetStream,XINE_PARAM_SPEED, XINE_SPEED_NORMAL);
     end
     else
     begin
        xine_set_param(myXStream.GetStream,XINE_PARAM_SPEED, XINE_SPEED_FAST_2);
     end
  end
end;

procedure TxineEngine.Seek (pos: Integer);
begin
    if( xine_get_stream_info( myXStream.GetStream, XINE_STREAM_INFO_SEEKABLE ) <> 0) then
    begin
        xine_play( myXStream.GetStream, 0, pos );
    end;
end;

procedure TxineEngine.Menu (id : Integer);
begin
end;

procedure TxineEngine.SetVolume(vol:Integer);
begin
     xine_set_param(myXStream.GetStream, XINE_PARAM_AUDIO_VOLUME, vol);
end;

function TxineEngine.GetAutoplayPluginList:TStringList;
var
     ids: PPChar;
begin
     ids := xine_get_autoplay_input_plugin_ids(myXine);
     result := PPCharToStringList(ids)
end;

function TxineEngine.OpenAutoplayInputPluginList(str: String):TStringList;
var
     mrls: PPChar;
     num: Integer;
begin
     mrls := xine_get_autoplay_mrls(myXine, PChar(str), @num);
     result := PPCharToStringList(mrls);
end;

procedure TxineEngine.Exit;
begin
     xine_exit(myXine);
end;

procedure TxineEngine.SetProperty (prop, value:Integer);
begin

     xine_set_param(myXStream.GetStream, prop, value);
end;

procedure TxineEngine.SetAudioChannel (chan: Integer);
begin
     xine_set_param(myXStream.GetStream, XINE_PARAM_AUDIO_CHANNEL_LOGICAL, chan);
end;

procedure TxineEngine.SetSubtitleChannel (chan: Integer);
begin
     xine_set_param(myXstream.GetStream,XINE_PARAM_SPU_CHANNEL,chan);
end;

procedure TxineEngine.SetZoom (zoom: Integer);
begin
     if (zoom < XINE_VO_ZOOM_MAX ) and (zoom > XINE_VO_ZOOM_MIN) then
     begin
          xine_set_param(myXStream.GetStream,XINE_PARAM_VO_ZOOM_X,zoom);
          xine_set_param(myXStream.GetStream,XINE_PARAM_VO_ZOOM_Y,zoom);
     end
     else
     begin
          if (zoom >= XINE_VO_ZOOM_MAX) then
          begin
               xine_set_param(myXStream.GetStream,XINE_PARAM_VO_ZOOM_X,
                    XINE_VO_ZOOM_MAX);
               xine_set_param(myXStream.GetStream,XINE_PARAM_VO_ZOOM_Y,
                    XINE_VO_ZOOM_MAX);
          end
          else
          begin
               xine_set_param(myXStream.GetStream,XINE_PARAM_VO_ZOOM_X,
                    XINE_VO_ZOOM_MIN);
               xine_set_param(myXStream.GetStream,XINE_PARAM_VO_ZOOM_Y,
                    XINE_VO_ZOOM_MIN);
          end
     end;
end;

procedure TxineEngine.SendInputEvent(button,x,y,typ: Integer);
var
     rect:  x11_rectangle_t;
     input: xine_input_data_t;
     event: xine_event_t;
begin
     rect.x := x;
     rect.y := y;
     rect.w := 0;
     rect.h := 0;

     xine_gui_send_vo_data(myXstream.GetStream,
          XINE_GUI_SEND_TRANSLATE_GUI_TO_VIDEO, @rect);

     input.x := x;
     input.y := y;
     input.button := button;

     event.typ := typ;
     event.data := @input;
     event.data_length := SizeOf(input);

     xine_event_send(myXStream.GetStream,@event);
end;

function TxineEngine.PPCharToStringList(items: PPChar):TStringList;
var
     i: Integer;
     list: TStringList;
begin
     i := 0;
     list := TStringList.Create;
     while (items <> nil) and (PCharArray(items)[i] <> nil) do
     begin
          list.Add(PCharArray(items)[i]);
          i := i + 1;
     end;
     result := list;
end;

constructor TxineEngine.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  myXine:=nil;
  Thread:=TMyThread.Create(true);
  Thread.xine:=self;
end;

destructor TxineEngine.Destroy;
begin
  if Assigned(myXine) then begin
    myVidPlug.vWidget:=nil;
    XLockDisplay(myVidPlug.Display);
    XUnmapWindow(myVidPlug.display, myVidPlug.xid);
    XUnlockDisplay(myVidPlug.display);
    XCloseDisplay (myVidPlug.display);
    xine_close(myXStream.GetStream);
    xine_event_dispose_queue(queue);
    xine_dispose(myXStream.GetStream);
    xine_close_audio_driver(myxine, myAudPlug.AoDriver);
    xine_close_video_driver(myxine, myVidPlug.VoDriver);
    xine_exit(myxine);
    myVidPlug.Destroy; myAudPlug.Destroy;
  end;
  inherited Destroy;
end;

procedure XineEventCB(user_data:Pointer; event: Pxine_event_t); cdecl;
var
     data: ^TxineUserData;
     handler: TXineEvent;
begin
     if (user_data <> NIL) then
     begin
          data := user_data;
          case event^.typ of
          XINE_EVENT_UI_PLAYBACK_FINISHED:
               begin
                    handler := data^.OnPlaybackFinished;
               end;
          XINE_EVENT_UI_CHANNELS_CHANGED:
               begin
                    handler := data^.OnChannelChanged;
               end;
          XINE_EVENT_UI_SET_TITLE:
               begin
                    handler := data^.OnSetTitle;
               end;
          XINE_EVENT_PROGRESS:
               begin
                    handler := data^.OnProgress;
               end;
          XINE_EVENT_UI_NUM_BUTTONS:
               begin
                    handler := data^.OnNumButtons;
               end;
          XINE_EVENT_MRL_REFERENCE:
               begin
                    handler := data^.OnMRLReference;
               end
          else
               begin
                    handler := NIL;
               end;
          end;
          if (Assigned(handler)) then
          begin
               handler(event^.data);
          end;
    end
end;

procedure TMyThread.test;
begin
//  writeln('TESTAAA');
  if v then
  begin
    xine.Open('/home/peter/Videos/Beverly_Hills_Cop_avi.flv');
    v:=false;
  end;
end;

{ TMyThread }
constructor TMyThread.Create(CreateSuspended: Boolean; const StackSize: SizeUInt);
begin
  v:=true;
  FreeOnTerminate := True;
  inherited Create(true, StackSize);
end; // TMyThread.Create

destructor TMyThread.Destroy;
begin
  inherited Destroy;
end;

procedure TMyThread.Execute;
begin
  repeat
    Synchronize(@Test);
  until (Terminated) or (Suspended);
  Suspend;
end;

end.
