{
                             xineVideoPlugin.pas
                             -------------------
    copyright            : (C) 2002 by Ryan Rusaw
    email                : rrusaw@sasktel.net

 ***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
}


unit xinevideoplugin;
{$DEFINE read_forward_definitions}
{$mode objfpc}{$H+}
interface

uses
  SysUtils, Classes, xine, xlib, Controls, x,Graphics,xvlib,xv,gtk,
  LCLIntf,LCLType, Forms,xatom,glib,libc,LCLClasses,GtkWSControls,gdk,
  WSLCLClasses, GTKDef,GtkInt, InterfaceBase, gdkpixbuf,xutil,pkghandler,
  GtkExtra,GtkWSPrivate,LCLProc,GTKProc,gtk2, Gtk2WSPrivate, Gtk2Def,
  gdk2,gdk2x,Gtk2Int;

var
  xvt: x11_visual_t;

const
     defaultXpos = 100;
     defaultYpos = 100;
     defaultWidth = 640;
     defaultHeight = 480;

{$define INPUT_MOTION (ExposureMask | ButtonPressMask | KeyPressMask | \
                      ButtonMotionMask | StructureNotifyMask |        \
                      PropertyChangeMask | PointerMotionMask)}
type
  {$IDefine GTKhasX}
  TvideoPlugins = (vdNone, vdDxr3, vdXv, vdXshm, vdAa, vdVidix, vdFb, vdSdl, vdAuto);

  { TxineVideoPlugin }

  TxineVideoPlugin = class(TComponent)
  private
    { Private declarations }
    vo_driver : Pxine_vo_driver_t;
    myVideoDriver : TvideoPlugins;
    myWidget : TWinControl;
    function getXID: TWindow;
    procedure setVideoDriver(driver: TvideoPlugins);
    procedure setViewWidget(w: TWinControl);
  protected
    { Protected declarations }
  public
    { Public declarations }
    property xid: TWindow read getXID;
    Display : pDisplay;
    Screen :Integer;
    ResH, ResV, DisplayRatio: Double;
    destructor Destroy; Override;

    property VoDriver : Pxine_vo_driver_t read vo_driver;
    procedure Initialize(Xine: Pxine_t; UsrDta: Pointer );
    function IsEmbeddable: Boolean;
  published
    { Published declarations }
    property videoDriver: TvideoPlugins read myVideoDriver
      write setVideoDriver default vdNone;
    property vWidget: TWinControl read myWidget write setViewWidget;
  end;

const
  vidNames : array[0..8] of PChar = ( 'none', 'dxr3', 'xv', 'xshm', 'aa',
            'vidix', 'fb', 'sdl', 'auto' );

procedure DestSizeCB ( data:Pointer; vid_w, vid_h:Integer;
       vid_pix_a : Double;
       dest_w, dest_h:PInteger;
       dest_pix_a : PDouble); cdecl;

procedure FrameOutputCB (data:Pointer; vid_w, vid_h:Integer;
       vid_pix_a: Double;
       dest_x, dest_y, dest_w, dest_h: Pinteger;
       dest_pix_a : PDouble;
       win_x, win_y:PInteger ); cdecl;

//function XShmGetEventBase(Display: PDisplay):Integer; cdecl;
//   external 'libXext.so';
//function XShmQueryExtension(Display: PDisplay):Integer; cdecl;
//   external 'libXext.so';

procedure Register;

implementation

uses
  xineengine;

const
  ShmCompletion = 1;

var
  UserData: TxineUserData;

function GetX11Window(AWidget: PGtkWidget): X.TWindow;
begin
// first make sure the x11 window is created for the gtk widget
if GTK_WIDGET_REALIZED(AWidget) = False then
gtk_widget_realize(AWidget);

// now retrieve the X window
{$IFDEF GTK2}
Result := GDK_WINDOW_XWINDOW(PGdkDrawAble(AWidget^.window));
{$ENDIF}
{$IFDEF GTK1}
Result := GDK_WINDOW_XWINDOW(PGdkWindowPrivate(AWidget^.window));
{$ENDIF}

end;

procedure Register;
begin
  RegisterComponents('Pluto', [TxineVideoPlugin]);
end;

function TxineVideoPlugin.IsEmbeddable: Boolean;
begin
{     if (myVideoDriver = vdXv) or (myVideoDriver = vdXshm) then
     begin}
          result := True;
{     end
     else
     begin
          result := False;
     end;}
end;

procedure TxineVideoPlugin.setViewWidget(w: TWinControl);
begin
     myWidget := w;
end;

destructor TxineVideoPlugin.Destroy;
begin
  myWidget:=nil;
  inherited Destroy;
end; // TxineVideoPlugin.Destroy

procedure TxineVideoPlugin.setVideoDriver(driver: TvideoPlugins);
begin
     myVideoDriver := driver;
end;

function TxineVideoPlugin.getXID: TWindow;
begin
  result:=GDK_WINDOW_XWINDOW(PGtkWidget(PtrUInt(myWidget.Handle))^.window);
  //GDK_WINDOW_XWINDOW(PGdkDrawable(PGdkWindowPrivate(PGtkWidget(myWidget.Handle)^.window)));
  //GDK_WINDOW_XWINDOW(PGdkDrawable(PGdkWindowPrivate(PGtkWidget(myWidget.Handle)^.window)));
end;


procedure TxineVideoPlugin.Initialize(Xine: Pxine_t; UsrDta: Pointer );
begin
  myWidget.Visible:=False;
     if (myWidget <> NIL) then begin
//       xid:=GDK_WINDOW_XWINDOW(PGdkDrawable(PGdkWindowPrivate(PGtkWidget(myWidget.Handle)^.window)));
//       XMapWindow(Display, xid);
     end
     else begin
     {     xid := XCreateSimpleWindow (Display,
			      XDefaultRootWindow(Display),
			      defaultXpos, defaultYpos,
			      defaultWidth, defaultHeight,
			      0, clBlack, clBlack);
          XMapWindow(Display, xid);}
     end;
     UserData := TxineUserData(UsrDta^);

     XSelectInput(Display,xid,
          ExposureMask or StructureNotifyMask);

     // Select Video Plugin
     xvt.display := Display;
     xvt.screen := Screen;
     xvt.d := xid;
     ResH := XDisplayWidth(Display,Screen)*1000
            / XDisplayWidthMM(Display,Screen);
     ResV := XDisplayHeight(Display,Screen)*1000
           / XDisplayHeightMM(Display,Screen);
     DisplayRatio := ResV/ResH;
        DisplayRatio := 1.0;
     if ((Abs(DisplayRatio)-1.0) < 0.01) then
     begin
        DisplayRatio := 1.0;
     end;
     xvt.dest_size_cb := @DestSizeCB;
     xvt.frame_output_cb := @FrameOutputCB;

     // Add to user_data record
     UserData.resizeFactor := 1.0;
     UserData.displayRatio := DisplayRatio;
     UserData.Display := Display;

     xvt.user_data := @UserData;

     vo_driver := xine_open_video_driver(Xine,vidNames[ord(myVideoDriver)],
       XINE_VISUAL_TYPE_X11, Pointer(@xvt));
end;

procedure DestSizeCB (data:Pointer; vid_w, vid_h:Integer; vid_pix_a : Double;
   dest_w, dest_h:PInteger; dest_pix_a: PDouble); cdecl;
var
   user_data: TxineUserData;
   temp : Double;
begin
  if Assigned(data) then
  begin
    user_data := TxineUserData(data^);
    if ( vid_pix_a >= user_data.displayRatio) then
    begin
       temp := vid_w*vid_pix_a;
       temp := temp/user_data.displayRatio+0.5;
       vid_w := Round(temp);
    end
    else
    begin
       temp := vid_h*user_data.displayRatio;
       temp := temp/vid_pix_a+0.5;
       vid_h := Round(temp);
    end;
    if ( user_data.resizeFactor = 0.0 ) then
    begin
       dest_w^ := user_data.rect.w;
       dest_h^ := user_data.rect.h;
    end
    else
    begin
       dest_w^ := Round(vid_w*user_data.resizeFactor);
       dest_h^ := Round(vid_h*user_data.resizeFactor);
    end;
    dest_pix_a^ := user_data.displayRatio;
  end;
end;

procedure FrameOutputCB (data:Pointer; vid_w, vid_h:Integer;
   vid_pix_a : Double; dest_x, dest_y, dest_w, dest_h: PInteger;
   dest_pix_a: PDouble;win_x, win_y : PInteger ); cdecl;
var
  temp : Double;
  user_data : TxineUserData;
  scale_x, scale_y : Double;
begin
   dest_pix_a^ :=2;
   vid_pix_a:=2;
  if Assigned(data) then begin
    user_data := TxineUserData(data^);
    if ( vid_pix_a >= user_data.displayRatio) then
    begin
       temp := vid_w*vid_pix_a;
       //temp := temp/user_data.displayRatio+0.5;
       vid_w := Round(temp);
    end
    else
    begin
       temp := vid_h*user_data.displayRatio;
     //  temp := temp/vid_pix_a+0.5;
       vid_h := Round(temp);
    end;
    if ( user_data.resizeFactor = 0.0 ) then
    begin
       dest_w^ := user_data.old_width;
       dest_h^ := user_data.old_height;
    end
    else
    begin
       vid_w := Round(vid_w*user_data.resizeFactor);
       vid_h := Round(vid_h*user_data.resizeFactor);
       scale_x := user_data.rect.w / vid_w;
       scale_y := user_data.rect.h / vid_h;

       if ( scale_x < scale_y ) then
       begin
          vid_w := Round(vid_w * scale_x);
          vid_h := Round(vid_h * scale_x);
          dest_x^ := -1 * (vid_w - user_data.rect.w) div 2;
          dest_y^ := -1 * (vid_h - user_data.rect.h) div 2;
       end
       else
       begin
          vid_w := Round(vid_w * scale_y);
          vid_h := Round(vid_h * scale_y);
          dest_x^ := -1 * (vid_w - user_data.rect.w) div 2;
          dest_y^ := -1 * (vid_h - user_data.rect.h) div 2;
       end;
       if (vid_w<>user_data.old_width) or (vid_h<>user_data.old_height) then
       begin
          user_data.old_width := vid_w;
          user_data.old_height := vid_h;
       end;
       dest_w^ := vid_w;
       dest_h^ := vid_h;
    end;
    user_data.resizeFactor := 0.1;
    dest_pix_a^ := user_data.displayRatio;
  end;
end;

end.
